"""
.. module:: QGA_performance_report.py
    :synopsis: Some functions to analyse and plot the performance of QGA
    compared to CGA.
.. moduleauthor::  Ruben Ibarrondo (rubenibarrondo@gmail.com)
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from QGA_vs_CGA import *
from time import time
from sklearn.linear_model import LinearRegression


def plot_pop_in_fitness(popevol, fitness_function, classical, n = None, cl = None, gpos=(0.9, 0.9)):
    fig = plt.figure()

    if classical:
        cl = len(popevol[0][0])
        n = len(popevol[0])
    else:
        if not cl or not n:
            raise Exception('For not classical version n and cl are required')

    xmax = 2**cl -1
    xmin = 0
    xx = np.arange(xmin, xmax+1)
    yy = np.array([fitness_function(binenc(x, cl)) for x in xx])
    ymax = max(yy)
    ymin = min(yy)

    def update_f(num):
        plt.cla()
        plt.grid(linestyle='dashed')
        plt.rc('axes', axisbelow=True)

        props = dict(boxstyle='round', facecolor='wheat', alpha=0.8)
        plt.text(xmin + (xmax - xmin) * gpos[0], ymin + (ymax - ymin) * gpos[1], 'g = {}'.format(num), bbox=props)
        plt.plot(xx, yy, color='tab:blue')

        if classical:
            cdic = {}
            for c in popevol[num]:
                cdic[c] = 1 + cdic.get(c, 0)
            for c, frequency in cdic.items():
                plt.plot(bindec(c), fitness_function(c), color='tab:orange', marker='o', markersize=30 * frequency / n)
        else:
            cdic = {}
            for tt in popevol[num]:
                pop, prob = tt[0], tt[1]
                for c in pop:
                    cdic[c] = prob + cdic.get(c, 0)

            for c, frequency in cdic.items():
                plt.plot(bindec(c), fitness_function(c), color='tab:orange', marker='o',
                         markersize=30 * frequency / n)

        # This is length specific!
        plt.xticks([0, 2, 4, 8, 16, 32, 64])

        plt.ylabel("$fitness$")
        plt.xlabel("individuals")

        plt.xlim([xmin - 0.05 * (xmax - xmin), xmax + 0.05 * (xmax - xmin)])
        plt.ylim([ymin - 0.05 * (ymax - ymin), ymax + 0.05 * (ymax - ymin)])

    anim = animation.FuncAnimation(fig, update_f, frames=len(popevol), interval=240)
    #anim.save("3_evol_fitnesMaxMean.gif")
    plt.show()


def get_fitnes_evol(popevol, fitness_function, classical):
    '''getting the quantum version of this, is not trivial, which is the sensible way?'''

    if classical:
        fitevol = []
        for generation in popevol:
            fitevol.append([fitness_function(c) for c in generation])
        return np.array(fitevol)
    else:
        fitevol = []
        for generation in popevol:
            fitgen = []
            for state in generation:
                fitgen.extend([fitness_function(c) for c in state[0]])
            fitevol.append(fitgen)
        return np.array(fitevol)


def plot_fitness_evol(popevol, fitness_function, classical):
    fv = get_fitnes_evol(popevol, fitness_function, classical)

    plt.plot([max(f) for f in fv], label='max')
    plt.plot([sum(f) / len(f) for f in fv], label='average')
    plt.plot([min(f) for f in fv], label='min')

    plt.legend()
    plt.grid(axis='y', linestyle='dashed')
    plt.show()


def get_first_time_podium(popevol, fitness_function, classical, tolerance=0):
    '''
    This function can be tricky for the quantum algorithm. We will focus on the first time
    one of the involved computational basis states went above the threshold.
    This may vary, maybe to take the probability of success into account.
    I will let you know.

    :param popevol:
    :param fitness_function:
    :param classical:
    :param tolerance:
    :return:
    '''
    if classical:
        cl = len(popevol[0][0])
    else:
        cl = len(popevol[0][0][0][0])

    xx = np.arange(0, 2**cl)
    yy = [fitness_function(binenc(x, cl)) for x in xx]
    yy.sort(reverse=True)
    threshold = yy[int(2**cl * tolerance)]

    first = None
    if classical:
        for g, pop in enumerate(popevol):
            if np.any([fitness_function(c) >= threshold for c in pop]):
                first = g
                break
    else:
        found = False
        for g, tuple_pop in enumerate(popevol):
            if found:
                break
            for pop, prob in tuple_pop:
                if np.any([fitness_function(c) >= threshold for c in pop]):
                    first = g
                    found = True
                    break
    return first


def unpack_qpopevol(popevol, n, cl):
    new_popevol = []
    for d in popevol:
        gen_d = []
        for i, j in d.keys():
            if i == j:
                ib = [binenc(i, cl * n)[index:index + cl] for index in range(0, n * cl, cl)]
                gen_d.append((ib, d.get((i, j))))
        new_popevol.append(gen_d)
    return new_popevol


def check_superposition_conservation(directory, nmax, states):
    for n in range(nmax):
        file = open(directory+'/'+'QGA_{:03d}'.format(n))
        init_range = list(range(9, 9+states+1))
        sorted_range = list(range(9 + 5 + 2*states, 9 + 5 + 3*states+1))
        for i, line in enumerate(file):
            if (i in init_range) or (i in sorted_range):
                print(line, end='')
        print()
        print('-'*32)
        print()
        file.close()


def classical_analysis(directory, trial_max, tolerance, plotshow=True, label=None):
    not_found = 0
    average = 0
    average_sq = 0

    first_arr = []
    initial_podium_fitness = []
    initial_best_fitness = []
    final_podium_fitness = []
    for trial in range(trial_max):
        if not label:
            popevol = parse_CGA_data(directory+'/'+'CGA_{:03d}'.format(trial))
        else:
            popevol = parse_CGA_data(directory+'/'+'CGA_{:03d}'.format(trial)+'_'+label)
        first = get_first_time_podium(popevol, parabolic, True, tolerance=tolerance)
        if first != None:
            average += first
            average_sq += first**2
            first_arr.append( - first)

            fitness = [parabolic(chromosome) for chromosome in popevol[0]]
            fitness.sort(reverse=True)
            initial_podium_fitness.append(sum(fitness[:4])/4)
            initial_best_fitness.append(fitness[0])

            fitness = [parabolic(chromosome) for chromosome in popevol[-1]]
            fitness.sort(reverse=True)
            final_podium_fitness.append(sum(fitness[:4]) / 4)

        else:
            first_arr.append(- 25)
            fitness = [parabolic(chromosome) for chromosome in popevol[0]]
            fitness.sort(reverse=True)
            initial_podium_fitness.append(sum(fitness[:4]) / 4)
            initial_best_fitness.append(fitness[0])

            fitness = [parabolic(chromosome) for chromosome in popevol[-1]]
            fitness.sort(reverse=True)
            final_podium_fitness.append(sum(fitness[:4]) / 4)

            not_found += 1
    print(not_found)
    print(average / (200 - not_found))
    print(math.sqrt(average_sq - average) / (200 - not_found))


    if plotshow:
        plt.plot(final_podium_fitness, first_arr, '.')
        #plt.plot(initial_best_fitness, first_arr, '.', alpha=0.5)

        plt.ylabel('first')
        plt.xlabel('final podium fitness')
        plt.xlim([1015, 1025])
        plt.show()


def quantum_analysis(directory, trial_max, tolerance, plotshow=True):
    not_found = 0
    average = 0
    average_sq = 0
    first_arr = []
    for trial in range(trial_max):
        popevol, n, cl, a = parse_QGA_data(directory+'/'+'QGA_{:03d}'.format(trial))
        popevol = unpack_qpopevol(popevol, n, cl)
        first = get_first_time_podium(popevol, parabolic, False, tolerance=tolerance)
        if first != None:
            average += first
            average_sq += first**2
            first_arr.append(- first)
        else:
            first_arr.append(- 25)
            not_found += 1
    print(not_found)
    print(average / (200 - not_found))
    print(math.sqrt(average_sq - average) / (200 - not_found))

    if plotshow:
        plt.plot(first_arr, '.')
        #plt.plot(initial_best_fitness, first_arr, '.', alpha=0.5)

        plt.ylabel('first')
        plt.xlabel('final podium fitness')
        plt.xlim([1015, 1025])
        plt.show()


def superposition_and_time_cost():
    fitness_oracle = lambda x, y: parabolic(x) < parabolic(y)

    G = 1
    N = 8
    cl = 6
    pm = 1/N/cl / 2 # one each 2 generations

    X = qm.rho([0, 0, 1, 1], [0, 1, 0, 1], [0, 1, 1, 0], (2, 2))

    dt2_arr = []
    dt3_arr = []
    dt2_dev = []
    dt3_dev = []
    sc_arr = []

    for superposition_coefficient in range(1, 21):
        print()
        print(superposition_coefficient, '|', end='')
        dt2 = 0
        dt2sq = 0
        dt3 = 0
        dt3sq = 0
        for trial in range(1):
            if trial % 10 == 0:
                print('*', end='')
            ii = [int(i) for i in np.random.randint(0, 2 ** (N * cl), superposition_coefficient)]
            coefficients = [float(c) for c in np.ones(superposition_coefficient) / np.sqrt(superposition_coefficient)]

            t1 = time()
            rho_population = qm.rho.gen_rho_from_state(coefficients, ii, (2 ** (N * cl)))
            t2 = time()
            QuantumGeneticAlgorithm(fitness_oracle, rho_population, generationNumber=G, pm=pm, N=N, cl=cl, mutation_unitary=X)
            t3 = time()

            dt2 += t2 - t1
            dt2sq += (t2 - t1)**2
            dt3 += t3 - t2
            dt3sq += (t3 - t2)**2

        sc_arr.append(superposition_coefficient)
        dt2_arr.append(dt2/200)
        dt2_dev.append(math.sqrt(dt2sq)/200)
        dt3_arr.append(dt3 / 200)
        dt3_dev.append(math.sqrt(dt3sq) / 200)

        if dt3/200 > 5:
            print('time limit exceded: ', dt3/200)
            break

    x = np.array(sc_arr).reshape((-1, 1))
    model = LinearRegression().fit(x, np.log10(np.array(dt3_arr)))
    dt3_arr_predict = model.predict(x)
    print()
    print('coefficient of determination:', model.score(x, np.log10(np.array(dt3_arr))))
    print('intercept:', model.intercept_)
    print('slope:', model.coef_)

    plt.plot(sc_arr, np.log10(np.array(dt3_arr)), '.')
    plt.plot(sc_arr, dt3_arr_predict, '--')
    plt.show()


def get_convergence_iteration(popevol, fitness_criteria):
    cg = None
    pmax_prev = None
    for g, population in enumerate(popevol):
        population = np.array(population)
        population_fitness = np.array([fitness_criteria(chromosome) for chromosome in population])
        pmax = population[np.where(population_fitness == max(population_fitness))][0]
        if g > 0:
            if pmax != pmax_prev:
                pmax_prev = pmax
                cg = g
        else:
            pmax_prev = pmax
            cg = g
    return cg


def deciding_convergence_generation(dirpath, fitness_function, trials):
    """
    Takes several CGAs and checks how many generations where needed for convergence.
    Realizes a plot to visualize the result.
    :return: None
    """

    cg_arr = []
    max_fit_arr = []
    for trial in range(trials):
        popevol = parse_CGA_data(dirpath+'/CGA_{:03d}'.format(trial))
        cg = get_convergence_iteration(popevol, fitness_function)
        cg_arr.append(cg)
        max_fit_arr.append(max([fitness_function(chromosome) for chromosome in popevol[cg]]))
    plt.hist(cg_arr)
    plt.show()

    N = len(cg_arr)
    cg_arr = np.array(cg_arr)
    freq_arr = []
    fit_arr = []
    for gen_lim in range(10, 110, 10):
        frequency = sum(1 for cg in cg_arr if cg < gen_lim)
        fitness = sum(max_fit_arr[cg] for cg in cg_arr if cg < gen_lim) / frequency
        print(gen_lim, frequency * 100 / N, fitness)
        freq_arr.append(frequency)
        fit_arr.append(fitness)

    fig, ax1 = plt.subplots()
    ax1.plot(freq_arr)
    ax2 = ax1.twinx()
    ax2.plot(fit_arr, color='C1')
    plt.show()

def get_final_population_above_threshold(tolerance=0, fitness_function=parabolic, cl=6, dirpath='GA_smart_reset'):
    xx = np.arange(0, 2 ** cl)
    yy = [fitness_function(binenc(x, cl)) for x in xx]
    yy.sort(reverse=True)
    threshold = yy[int(2 ** cl * tolerance)]

    above_threshold_stats = 0
    for trial in range(200):
        popevol, n, cl, a = parse_QGA_data(dirpath+'/QGA_{:03d}'.format(trial))
        popevol = unpack_qpopevol(popevol, n, cl)

        prob_above_threshold = 0
        for population, probability in popevol[-1]:
            for chromosome in population:
                if fitness_function(chromosome) >= threshold:
                    prob_above_threshold += probability
                    break
        above_threshold_stats += prob_above_threshold
    q_above_treshold_stats = above_threshold_stats / 200


    above_threshold_stats = 0
    for trial in range(200):
        popevol = parse_CGA_data(dirpath+'/CGA_{:03d}'.format(trial))

        prob_above_threshold = 0
        for chromosome in popevol[-1]:
            if fitness_function(chromosome) >= threshold:
                prob_above_threshold += 1
                break
        above_threshold_stats += prob_above_threshold
    c_above_treshold_stats = above_threshold_stats / 200

    return q_above_treshold_stats, c_above_treshold_stats


def get_reference_chromosome_in_final_population(reference_chromosome, dirpath):
    ref_chrom_stats = 0
    for trial in range(200):
        popevol, n, cl, a = parse_QGA_data(dirpath + '/QGA_{:03d}'.format(trial))
        popevol = unpack_qpopevol(popevol, n, cl)

        prob_ref_chrom = 0
        for population, probability in popevol[-1]:
            for chromosome in population:
                if chromosome == reference_chromosome:
                    prob_ref_chrom += probability
                    break
        ref_chrom_stats += prob_ref_chrom
    q_ref_chrom_stats = ref_chrom_stats / 200

    ref_chrom_stats = 0
    for trial in range(200):
        popevol = parse_CGA_data(dirpath + '/CGA_{:03d}'.format(trial))

        prob_ref_chrom = 0
        for chromosome in popevol[-1]:
            if chromosome == reference_chromosome:
                prob_ref_chrom += 1
                break
        ref_chrom_stats += prob_ref_chrom
    c_ref_chrom_stats = ref_chrom_stats / 200

    return q_ref_chrom_stats, c_ref_chrom_stats


if __name__ == '__main__':


    fitness_function = saw_teeth
    dirpath = 'GA_saw_teeth'
    others_q = 1
    others_c = 1
    for i in [19, 29, 39, 59]:
        qth, cth = get_reference_chromosome_in_final_population(binenc(i, 6), dirpath)
        print('{:4d}    {:6.2f}    {:6.2f}'.format(i, qth * 100, cth * 100))
        others_q -= qth
        others_c -= cth
    print('others  {:6.2f}    {:6.2f}'.format(others_q * 100, others_c * 100))

    print()
    print('Normalized to maximum converged')
    for i in [19, 29, 39, 59]:
        qth, cth = get_reference_chromosome_in_final_population(binenc(i, 6), dirpath)
        print('{:4d}    {:6.2f}    {:6.2f}'.format(i, qth * 100 / (1 - others_q), cth * 100 / (1 - others_c)))

