"""
.. module:: QGA_vs_CGA.py
    :synopsis: Implements the QGA based on different registers which are pseudo cloned and swapped to implement
    crossover, and semi-sorted to implement selection. CGA based os similar characteristics is also implemented for
    comparision.
    This GA model is only valid when the fitness can be evaluated by pairs.
.. moduleauthor::  Ruben Ibarrondo (rubenibarrondo@gmail.com)
"""

import numpy as np
import math
import quantum_mats as qm
from copy import deepcopy
from time import time

def ClassicGeneticAlgorithm(fitness_oracle, init_population,
                            generationNumber=100, pm=0.001,
                            store_path=None, fitness_function=None):
    # population number
    N = len(init_population)
    # chromosome length
    cl = len(init_population[0])
    population = deepcopy(init_population)
    if store_path:
        store = open(store_path, 'w')

    # 1 - Get the initial population
    generation = 0
    while generation < generationNumber:
        if store_path:
            print('-' * 32, '\n',
                  ' ' * 9, 'Generation: {}'.format(generation),
                  '\n', '-' * 32, '\n',
                  file=store, sep='')
            print('initial population:\n', ' '.join(''.join(str(c) for c in p) for p in population), file=store)
            print('fitness:\n', ' '.join([str(fitness_function(p)) for p in population]), file=store)
            print(file=store)

        generation += 1
        # 2 - sort or semi sort,
        # we will perform it manually to ensure the procedure is similar to the quantum one.
        # Set sort stage limit to N-1 to get a full sort.
        # Set sort stage limit to N/2 to get a partial sort.
        for sort_stage in range(N-1):
            for comp in range(sort_stage % 2, N, 2):
                if comp+1 < N:
                    swap = fitness_oracle(population[comp], population[comp+1])
                    if swap:
                        temp = population[comp]
                        population[comp] = population[comp+1]
                        population[comp+1] = temp
        if store_path:
            print('sorted population:\n', ' '.join(''.join(str(c) for c in p) for p in population), file=store)
            print('fitness:\n', ' '.join([str(fitness_function(p)) for p in population]), file=store)
            print(file=store)

        # 3 - Clear the lowest half
        for i in range(N):
            if i < N/2:
                population[i] = population[i]
            else:
                population[i] = [0 for c in population[i]]
        if store_path:
            print('population distribution after sort and clear:\n', ' '.join(''.join(str(c) for c in p) for p in population), file=store)
            print('fitness:\n', ' '.join([str(fitness_function(p)) for p in population]), file=store)
            print(file=store)

        # 4 - Crossover to fill
        # we will cross by pairs, and choose a central crossing index provisionally
        for c in range(0, N//2, 2):
            cross_index = cl//2
            population[c+N//2] = population[c][:cross_index] + population[c+1][cross_index:]
            population[c+N//2+1] = population[c+1][:cross_index] + population[c][cross_index:]
        if store_path:
            print('population after crossover:\n', ' '.join(''.join(str(c) for c in p) for p in population), file=store)
            print('fitness:\n', ' '.join([str(fitness_function(p)) for p in population]), file=store)
            print(file=store)

        # 5 - mutate with probability pm
        for p in population:
            for c in range(cl):
                r = np.random.random()
                if r < pm:
                    p[c] = (p[c]+1) % 2
        if store_path:
            print('population after mutation:\n', ' '.join(''.join(str(c) for c in p) for p in population), file=store)
            print('fitness:\n', ' '.join([str(fitness_function(p)) for p in population]), file=store)
            print(file=store)
    if store_path:
        store.close()
    return population


# -----------------------------
#
#       QUANTUM VERSION
#
# -----------------------------


def QuantumGeneticAlgorithm(fitness_oracle, init_population,
                            generationNumber=100, pm=0.001,
                            N = None, cl = None,
                            mutation_unitary = None, pre_projection_rotation = 1,
                            store_path=None, fitness_function=None):
    """
    Gets the same arguments and then it passes them to density matrices and unitary matrices.
    The building of the fitness_oracle could be optimized.
    """

    # 1 - Get the initial population
    # add the ancilla states!
    if type(init_population) == qm.rho:
        if not N or not cl:
            raise ValueError('If population is a quantum_matrix.rho object N and cl have to be defined.')
        a = N//2
        ancillas = [ia for ia in range(N*cl, N*cl+a)]
        rho_pop = qm.rho([], [], [], (init_population.shape[0] * 2**a, init_population.shape[1] * 2**a))
        for i, j in init_population.store.keys():
            rho_pop.set(i * 2**a, j * 2**a, init_population.get(i, j))
    else:
        population = deepcopy(init_population)
        # population number
        N = len(population)
        # chromosome length
        cl = len(population[0])
        i = 0
        index = 0
        for p in population:
            for c in p:
                i += c * 2**(N*cl - index - 1)
                index += 1
        a = N//2
        ancillas = [ia for ia in range(N*cl, N*cl+a)]
        rho_pop = qm.rho([1], [i * 2**a], [i * 2**a], (2**(N*cl + a), 2**(N*cl + a)))

    if store_path:
        store = open(store_path, 'w')
        print('Genetic Algorithm parameters:', file=store)
        print('generationNumber = {:d}, pm = {:f}, population number = {:d}, chromosome length = {:d}'.format(
            generationNumber, pm, N, cl), file=store)
        print(file=store)

    # Build de cloning routine unitary
    clone = qm.Identity(2 ** (N * cl))
    for nu in range(0, N // 2):
        s = qm.Swap_reg(range((nu + 1) * cl, (nu + 2) * cl),
                        range((nu + N // 2) * cl, (nu + 1 + N // 2) * cl),
                        N * cl)
        u = qm.Uclone(2 ** cl)
        unu = s.dot(qm.KronExpand(2 ** (nu * cl), u, 2 ** ((N - nu - 2) * cl))).dot(s)
        clone = clone.dot(unu)
    clone = qm.KronExpand(1, clone, 2 **(N // 2))

    # Build the crossover routine unitary
    cross_index = cl // 2
    qq1 = [q1 for nu in range(0, N//2, 2) for q1 in range((N // 2 + nu) * cl + cross_index, (N // 2 + nu + 1) * cl)]
    qq2 = [q2 for nu in range(0, N//2, 2) for q2 in range((N // 2 + nu + 1) * cl + cross_index, (N // 2 + nu + 2) * cl)]
    cross = qm.Swap_reg(qq1, qq2, N*cl)
    cross = qm.KronExpand(1, cross, 2 ** (N // 2))

    # Build the sort/semi sort step unitary
    # You cant build it completely because you need to set the ancillas to
    # zero in each stage, but this makes it easier latter.
    sort_arr = [qm.Identity(2 ** (N * cl + N // 2)), qm.Identity(2 ** (N * cl + N // 2))]
    for stage in [0, 1]:
        for mu in range(0, N - 1, 2):
            nu = mu + stage % 2
            if nu + 2 > N:
                continue
            orac = qm.Oracle(fitness_oracle, N * cl + N // 2,
                             range(nu * cl, (nu + 1) * cl),
                             range((nu + 1) * cl, (nu + 2) * cl), N * cl + mu // 2)
            cs = qm.CSwap_reg(N * cl + nu // 2,
                              range(nu * cl, (nu + 1) * cl),
                              range((nu + 1) * cl, (nu + 2) * cl),
                              N * cl + N // 2)
            sort_arr[stage] = sort_arr[stage].dot(cs).dot(orac)
    sort0, sort1 = sort_arr

    # Build mutation arrays
    X = qm.rho([1, 1], [0, 1], [1, 0], (2, 2))
    if not mutation_unitary:
        Xarr = []
        for i in range(N*cl):
            Xarr.append(qm.KronExpand(2**i, X, 2**(N*cl-i-1+a)))
    else:
        if type(mutation_unitary) != qm.rho:
            raise ValueError('mutation_unitary has to be a quantum_matrix.rho object, not {}.'.format(type(mutation_unitary)))
        Xarr = []
        for i in range(N * cl):
            Xarr.append(qm.KronExpand(2 ** i, mutation_unitary, 2 ** (N * cl - i - 1 + a)))


    generation = 0
    while generation < generationNumber:
        if store_path:
            print('-'*32, '\n',
                  ' '*9, 'Generation: {}'.format(generation),
                  '\n', '-' * 32, '\n',
                  file=store, sep='')
            print('initial state, population + ancillas:\n', rho_pop.store, file=store)
            population = []
            for state in rho_pop.str_qubit_state(block=cl).split('\n'):
                pop1, pop2 = state.split(', ')
                if pop1 == pop2.split(':')[0]:
                    pop = pop1
                else:
                    continue
                pop = pop.split()
                pop.pop(-1)
                prob = float(state.split(': ')[-1])
                population.append((pop, prob))
            i = 0
            print('initial population distribution:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join(pop),
                      file=store)
                i += 1
            i = 0
            print('fitness:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join([str(fitness_function(p)) for p in pop]),
                      file=store)
                i += 1
            print(file=store)
        generation += 1

        # 2 - sort or semi sort:
        #     N//2 sets semisort
        #     N sets sort
        for stage in range(0, N):
            sort = [sort0, sort1][stage % 2]
            rho_pop = sort.dot(rho_pop).dot(qm.Transpose(sort))
            for ai in ancillas:
                # This is what i call smart reset
                rho_pop.projection_controlled_rotation(ai, 1, X, pre_projection_rotation)

                # This is simple reset
                # rho_pop.set_qubit(ai, 'random')
        if store_path:
            print('sorted state, population + ancillas:\n', rho_pop.store, file=store)
            population = []
            for state in rho_pop.str_qubit_state(block=cl).split('\n'):
                pop1, pop2 = state.split(', ')
                if pop1 == pop2.split(':')[0]:
                    pop = pop1
                else:
                    continue
                pop = pop.split()
                pop.pop(-1)
                prob = float(state.split(': ')[-1])
                population.append((pop, prob))
            i = 0
            print('sorted population distribution:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join(pop),
                      file=store)
                i += 1
            i = 0
            print('fitness:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join([str(fitness_function(p)) for p in pop]),
                      file=store)
                i += 1
            print(file=store)


        # 3 - Clear the lowest half
        for q in range(N//2 * cl, N * cl):
            # This is what i call smart reset
            rho_pop.projection_controlled_rotation(q, 1, X, pre_projection_rotation)

            # This is simple reset
            # rho_pop.set_qubit(q, 'random')
        if store_path:
            print('state after sort and clear, population + ancillas:\n', rho_pop.store, file=store)
            population = []
            for state in rho_pop.str_qubit_state(block=cl).split('\n'):
                pop1, pop2 = state.split(', ')
                if pop1 == pop2.split(':')[0]:
                    pop = pop1
                else:
                    continue
                pop = pop.split()
                pop.pop(-1)
                prob = float(state.split(': ')[-1])
                population.append((pop, prob))
            i = 0
            print('population distribution after sort and clear:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join(pop),
                      file=store)  # this should be computed
                i += 1
            i = 0
            print('fitness:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join([str(fitness_function(p)) for p in pop]),
                      file=store)  # this should be computed
                i += 1
            print(file=store)


        # 4 - Crossover to fill
        rho_pop = clone.dot(rho_pop).dot(qm.Transpose(clone))
        rho_pop = cross.dot(rho_pop).dot(qm.Transpose(cross))
        if store_path:
            print('state after crossover, population + ancillas:\n', rho_pop.store, file=store)
            population = []
            for state in rho_pop.str_qubit_state(block=cl).split('\n'):
                pop1, pop2 = state.split(', ')
                if pop1 == pop2.split(':')[0]:
                    pop = pop1
                else:
                    continue
                pop = pop.split()
                pop.pop(-1)
                prob = float(state.split(': ')[-1])
                population.append((pop, prob))
            i = 0
            print('population after crossover:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join(pop),
                      file=store)
                i += 1
            i = 0
            print('fitness:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join([str(fitness_function(p)) for p in pop]),
                      file=store)
                i += 1
            print(file=store)

        # 5 - mutate with probability pm
        for c in range(N*cl):
            r = np.random.random()
            if r < pm:
                rho_pop = Xarr[c].dot(rho_pop).dot(Xarr[c])
        if store_path:
            print('state after mutation, population + ancillas:\n', rho_pop.store, file=store)
            population = []
            for state in rho_pop.str_qubit_state(block=cl).split('\n'):
                pop1, pop2 = state.split(', ')
                if pop1 == pop2.split(':')[0]:
                    pop = pop1
                else:
                    continue
                pop = pop.split()
                pop.pop(-1)
                prob = float(state.split(': ')[-1])
                population.append((pop, prob))
            i = 0
            print('population after mutation:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join(pop),
                      file=store)
                i += 1
            i = 0
            print('fitness:', file=store)
            for pop, prob in population:
                print('{:3d}: prob = {:5.3f}: '.format(i, prob),
                      ' '.join([str(fitness_function(p)) for p in pop]),
                      file=store)
                i += 1
            print(file=store)
    if store_path:
        store.close()
    return rho_pop


def parse_CGA_data(filepath):
    ''''
    Only interested in the initial population and the final population in each generation.
    '''
    file = open(filepath)
    popevol = []
    g = 0
    key_word = ''
    for line in file:
        if g == 0 and 'initial population' in line:
            key_word = 'initial population'
            g += 1
            continue

        if key_word == 'initial population':
            popevol.append(line.strip().split())
            key_word = ''

        if key_word == 'population after mutation':
            popevol.append(line.strip().split())

        if 'population after mutation' in line:
            key_word = 'population after mutation'
            continue
        else:
            key_word = ''
    file.close()
    return popevol


def parse_QGA_data(filepath, remove_ancillas = True):
    ''''
    Only interested in the initial state and the final state in each generation.
    The number of registers (n), register length (cl) and amount of ancillas (a),
    are also returned. If remove_ancillas is activated (defoult) ancilass are removed.
    :returns popevol, n, cl, a
    '''
    file = open(filepath)
    popevol = []

    g = 0
    key_word = ''
    for line in file:
        if g == 0 and 'initial state' in line:
            key_word = 'initial state'
            g += 1
            continue

        if key_word == 'initial state':
            popevol.append(eval(line.strip()))
            key_word = ''

        if g == 1 and 'initial population distribution' in line:
            key_word = 'initial population distribution'
            g += 1
            continue

        if key_word == 'initial population distribution':
            popstr = line.strip().split(': ')[-1]
            pop = popstr.split()
            n = len(pop)
            cl = len(pop[0])
            a = n // 2

        if key_word == 'state after mutation':
            popevol.append(eval(line.strip()))

        if 'state after mutation' in line:
            key_word = 'state after mutation'
            continue
        else:
            key_word = ''
    file.close()

    if remove_ancillas:
        new_popevol = []
        for pop in popevol:
            new_pop = {}
            for k, v in pop.items():
                i, j = k
                new_pop[(int(i // 2**a), int(j // 2**a))] = v
            new_popevol.append(new_pop)
        popevol = new_popevol
    return popevol, n, cl, a


def binenc(i, nq):
    ib = bin(i)[2:]
    if len(ib) < nq:
        ib = '0' * (nq - len(ib)) + ib
    return ib


def bindec(ib):
    return sum(2 ** (len(ib) - j - 1) * int(ib[j]) for j in range(len(ib)))


def parabolic(ib):
    return 2**(2*len(ib)-2) - (2**(len(ib)-1) - sum(2 ** (len(ib) - j - 1) * int(ib[j]) for j in range(len(ib))))**2


def saw_teeth(ib):
    x = sum(2 ** (len(ib) - j - 1) * int(ib[j]) for j in range(len(ib)))
    return x % 20 + x % 30

def simillar_pop(population_a):
    population_b = []
    first = True
    for individual in population_a:
        if first:
            i = bindec(''.join(str(chromosome) for chromosome in individual))
            jb = binenc(int(2 ** cl - i), cl)
            population_b.append([int(chromosome) for chromosome in jb])
            first = False
        else:
            population_b.append(individual.copy())
    return population_b


if __name__ == '__main__':
    fitness_oracle = lambda x, y: saw_teeth(x) < saw_teeth(y)

    G = 100
    N = 8
    cl = 6
    pm = 1 / N / cl / 4

    for trial in range(0, 200):
        population_a = []
        for p in range(N):
            population_a.append(list(np.random.randint(0, 2, cl)))

        ClassicGeneticAlgorithm(fitness_oracle, population_a, generationNumber=G, pm=pm,
                                store_path='GA_saw_teeth_convergence/CGA_{:03d}'.format(trial), fitness_function=parabolic)

    assert False, 'THE END'

    fitness_oracle = lambda x, y: saw_teeth(x) < saw_teeth(y)

    G = 50
    N = 8
    cl = 6
    pm = 1 / N / cl / 4

    theta = math.pi / 4
    H = qm.rho([math.cos(theta), -math.cos(theta), math.sin(theta), math.sin(theta)], [0, 1, 0, 1], [0, 1, 1, 0], (2, 2))

    for trial in range(0, 200):
        print(trial, end=' ')
        t1 = time()
        population_a = []
        for p in range(N):
            population_a.append(list(np.random.randint(0, 2, cl)))

        ClassicGeneticAlgorithm(fitness_oracle, population_a, generationNumber=G, pm=pm,
                                store_path='GA_saw_teeth/CGA_{:03d}'.format(trial), fitness_function=parabolic)

        ia = 0
        index = 0
        for p in population_a:
            for c in p:
                ia += c * 2 ** (N * cl - index - 1)
                index += 1

        rho_population = qm.rho([1], [int(ia)], [int(ia)], (2 ** (N * cl), 2 ** (N * cl)))
        QuantumGeneticAlgorithm(fitness_oracle, rho_population, generationNumber=G, pm=pm, N=N, cl=cl,
                                mutation_unitary=H, pre_projection_rotation=H,
                                store_path='GA_saw_teeth/QGA_{:03d}'.format(trial), fitness_function=parabolic)
        print(time() - t1)
